/******************************************************************************

Comments:
  This code is based on the demo.cpp example program from NVIDIA.
  It has been extended by Russ Taylor to handle vertex-plus-fragment
  code raytracing of Non-Polygonal objects, in particular implicit surfaces, in
  particular quadratics, in particular spheres.

******************************************************************************/

/******************************************************
 * Menagerie Demo:

 *
 ******************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef WIN32
# include <windows.h>
#endif
#ifdef __APPLE__
# include <OpenGL/gl.h>
# include <OpenGL/glext.h>
# include <GLUT/glut.h>
#else
# include <GL/gl.h>
# include <GL/glut.h>
# include <GL/glext.h>
# include <sys/types.h>
# include <sys/timeb.h>
#endif
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include "ppm.h"

#ifndef CWD
# define CWD ""
#endif

// Define this to put texture on the objects.
//#define TEXTURE

int g_curFrameNum = 0;

const char g_vertex_program_name[] = CWD "NPO_vertex.cg";
#ifdef TEXTURE
const char g_fragment_program1_name[] = CWD "NPO_tex_sphere_fragment.cg";
#else
const char g_fragment_program1_name[] = CWD "NPO_sphere_fragment.cg";
#endif
const char g_fragment_program2_name[] = CWD "NPO_louverblind_fragment.cg";
const char g_fragment_program3_name[] = CWD "NPO_cone_fragment.cg";
const char g_fragment_program4_name[] = CWD "NPO_nestsphere_fragment.cg";


// Global variables: hold the Cg context that we're storing our programs
// in as well as handles to the vertex and fragment program used in this
// demo.

CGcontext context;
CGprogram vertexProgram, fragmentProgram1, fragmentProgram2, fragmentProgram3, fragmentProgram4, fragmentProgram5;
CGprofile vertexProfile, fragmentProfile;

// More globals
const double g_hitherDist = 1;
const double g_yonDist = 200.0;
static bool g_stopMotion = false;
static bool g_showPolies = false;     // Show bounding geometry rather than Cg
static bool g_toggle = false;	      // Toggle back and forth between polygons and Cg
static unsigned g_whichMobile = 0;    // Which mobile to draw

static	GLuint	g_ballTexture;	      // Texture name for the texture on the ball.
static	GLuint	g_earthTexture;	      // Texture name for the texture on the earth in the mobile
static	GLuint	g_snowTexture;	      // Texture name for the texture on the snow ball
static	GLuint	g_rocketTexture;      // Texture name for the texture on the rocket in the mobile

// Forward declarations of functions

static void keyboard(unsigned char key, int x, int y);
static void display();
static void handleCgError();
static void DrawRocket();
static void DrawEarth();
static void DrawSnowman();
static void DrawMobile();
static void DrawSphereNest();
static void DrawRoom();
static void DrawBouncingBall();
static void UnitCube();
static void LoadCgPrograms();
static void ChooseProfiles();

// NOTE: When these get bound, they set the modelview matrix and others
// to the current one on the stack -- this makes any modeling transform
// that happens in the display list of the object have no effect!!!!!
// This means that we can't do vertex-list primitives and use the
// model transform.
inline void BindVertexParameters()
{
  static bool initialized = false;
  static CGparameter ModelViewProj, ModelViewI, ModelViewIT, ModelView;

  if (!initialized) {
    ModelViewProj = cgGetNamedParameter(vertexProgram, "ModelViewProj");
    ModelViewI = cgGetNamedParameter(vertexProgram, "ModelViewI");
    initialized = true;
  }

  cgGLSetStateMatrixParameter(ModelViewProj,
			      CG_GL_MODELVIEW_PROJECTION_MATRIX,
			      CG_GL_MATRIX_IDENTITY);
  cgGLSetStateMatrixParameter(ModelViewI,
			      CG_GL_MODELVIEW_MATRIX,
			      CG_GL_MATRIX_INVERSE);
}

void  LoadAndCreateTextures(void)
{
  int i,j;
  int r,g,b;
  const	int texture_res = 256;

  // Read the image file for the ball and resample to texture_res by texture_res.
  // We need to map the whole range (-1..1) in X and Y from the PPM file into the
  // texture.
  static  GLubyte ballImage[texture_res][texture_res][4];
  PPM ballPPM("ball_texture.ppm");
  for (i = 0; i < texture_res; i++) {
    for (j = 0; j < texture_res; j++) {
      ballPPM.Value_at_normalized( -1 + (2.0*i)/texture_res, -1 + (2.0*j)/texture_res, &r, &g, &b);
      ballImage[j][i][0] = r;
      ballImage[j][i][1] = g;
      ballImage[j][i][2] = b;
      ballImage[j][i][3] = 1;
    }
  }

  // Create the ball texture based on this buffer.
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &g_ballTexture);
  glBindTexture(GL_TEXTURE_2D, g_ballTexture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texture_res,
	      texture_res, 0, GL_RGBA, GL_UNSIGNED_BYTE,
	      ballImage);

  // Read the image file for the Earth and resample to texture_res by texture_res.
  // We need to map the whole range (-1..1) in X and Y from the PPM file into the
  // texture.
  static  GLubyte earthImage[texture_res][texture_res][4];
  PPM earthPPM("clear_earth.ppm");
  for (i = 0; i < texture_res; i++) {
    for (j = 0; j < texture_res; j++) {
      earthPPM.Value_at_normalized( -1 + (2.0*i)/texture_res, -1 + (2.0*j)/texture_res, &r, &g, &b);
      earthImage[j][i][0] = r;
      earthImage[j][i][1] = g;
      earthImage[j][i][2] = b;
      earthImage[j][i][3] = 1;
    }
  }

  // Create the Earth texture based on this buffer.
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &g_earthTexture);
  glBindTexture(GL_TEXTURE_2D, g_earthTexture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texture_res,
	      texture_res, 0, GL_RGBA, GL_UNSIGNED_BYTE,
	      earthImage);

  // Read the image file for the Snowball and resample to texture_res by texture_res.
  // We need to map the whole range (-1..1) in X and Y from the PPM file into the
  // texture.
  static  GLubyte snowImage[texture_res][texture_res][4];
  PPM snowPPM("snow_ball.ppm");
  for (i = 0; i < texture_res; i++) {
    for (j = 0; j < texture_res; j++) {
      snowPPM.Value_at_normalized( -1 + (2.0*i)/texture_res, -1 + (2.0*j)/texture_res, &r, &g, &b);
      snowImage[j][i][0] = r;
      snowImage[j][i][1] = g;
      snowImage[j][i][2] = b;
      snowImage[j][i][3] = 1;
    }
  }

  // Create the snow texture based on this buffer.
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &g_snowTexture);
  glBindTexture(GL_TEXTURE_2D, g_snowTexture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texture_res,
	      texture_res, 0, GL_RGBA, GL_UNSIGNED_BYTE,
	      snowImage);

  // Read the image file for the rocket and resample to texture_res by texture_res.
  // We need to map the whole range (-1..1) in X and Y from the PPM file into the
  // texture.
  static  GLubyte rocketImage[texture_res][texture_res][4];
  PPM rocketPPM("rocket.ppm");
  for (i = 0; i < texture_res; i++) {
    for (j = 0; j < texture_res; j++) {
      rocketPPM.Value_at_normalized( -1 + (2.0*i)/texture_res, -1 + (2.0*j)/texture_res, &r, &g, &b);
      rocketImage[j][i][0] = r;
      rocketImage[j][i][1] = g;
      rocketImage[j][i][2] = b;
      rocketImage[j][i][3] = 1;
    }
  }

  // Create the ball texture based on this buffer.
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &g_rocketTexture);
  glBindTexture(GL_TEXTURE_2D, g_rocketTexture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texture_res,
	      texture_res, 0, GL_RGBA, GL_UNSIGNED_BYTE,
	      rocketImage);

  // Other texture stuff we may need...
  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
}

///////////////////////////////////////////////////////////////////////////
// Main program; do basic GLUT and Cg setup, but leave most of the work
// to the display() function.

int main(int argc, char *argv[]) {
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_ALPHA | GLUT_DEPTH);
    glutInitWindowSize(512,512);
    glutCreateWindow("Menagerie");

    glutKeyboardFunc(keyboard);
    glutDisplayFunc(display);

    // Basic Cg setup; register a callback function for any errors
    // and create an initial context
    cgSetErrorCallback(handleCgError);
    context = cgCreateContext();

    // Set up Cg programs
    static bool haveLoadedPrograms = false;
    if (!haveLoadedPrograms) {
	// Do one-time setup only once; setup Cg programs and textures
	// and set up OpenGL state.
        ChooseProfiles();
        LoadCgPrograms();

        glEnable(GL_DEPTH_TEST);
        haveLoadedPrograms = true;
    }

    glShadeModel(GL_FLAT);

#ifdef	TEXTURE
    LoadAndCreateTextures();
#endif

    // Now make sure that the vertex and fragment programs, loaded
    // in LoadCgPrograms(), are bound.
    cgGLEnableProfile(vertexProfile);
    cgGLEnableProfile(fragmentProfile);
    cgGLBindProgram(vertexProgram);
    cgGLBindProgram(fragmentProgram1);

    // Now bind uniform parameters to vertex program
    cgGLSetParameter1f(cgGetNamedParameter(vertexProgram, "Zhither"), g_hitherDist);
    cgGLSetParameter1f(cgGetNamedParameter(vertexProgram, "Zyon"), g_yonDist);

    // Now bind uniform parameters to fragment programs
    cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram1, "Zhither"), g_hitherDist);
    cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram1, "Zyon"), g_yonDist);
    cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram2, "Zhither"), g_hitherDist);
    cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram2, "Zyon"), g_yonDist);
    cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram3, "Zhither"), g_hitherDist);
    cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram3, "Zyon"), g_yonDist);
    cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram4, "Zhither"), g_hitherDist);
    cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram4, "Zyon"), g_yonDist);

    // and all the rest happens in the display function...
    glutMainLoop();

    return 0;
}

// display callback function

static void display() 
{
    // The usual OpenGL stuff to clear the screen and set up viewing.
    //glClearColor(.25, .25, .25, 1.);
    glClearColor(0.0,0.0,0.0, 1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(40., 1.0, g_hitherDist, g_yonDist);

    glMatrixMode(GL_MODELVIEW);

    glLoadIdentity();
    gluLookAt(0, 0, 30, 0, 0, 0, 0, 1, 0);

    // Set things up so that backface culling is on.
    glCullFace(GL_BACK);
    glEnable(GL_CULL_FACE);

    // Disable the rendering profile, draw the room, then re-enable it.
    cgGLDisableProfile(vertexProfile);
    cgGLDisableProfile(fragmentProfile);
    glPushMatrix();
      DrawRoom();
    glPopMatrix();
    if (!g_showPolies) {
      cgGLEnableProfile(vertexProfile);
      cgGLEnableProfile(fragmentProfile);
    }

    switch (g_whichMobile % 3) {
    case 0:
      // Draw the rocket ship and Earth mobile.
      glPushMatrix();
	DrawMobile();
      glPopMatrix();
      break;

    case 1:
      // Draw a nested sphere
      glPushMatrix();
	DrawSphereNest();
      glPopMatrix();
      break;
    }

    glPushMatrix();
      DrawSnowman();
    glPopMatrix();

    // Draw a bouncing ball
    glPushMatrix();
      DrawBouncingBall();
    glPopMatrix();

    glutSwapBuffers();

    // Force another display(), so that the objects keep moving
    glutPostRedisplay();

    // If we are toggling, then switch polies on and off periodically
    if (g_toggle) {
      static  int tCount = 0;
      if (++tCount == 79) {
	tCount = 0;
	g_showPolies = !g_showPolies;
      }
    }

    static struct _timeb start;
    if (g_curFrameNum == 0) {
      _ftime( &start );
    }
    if (!g_stopMotion) {
      g_curFrameNum++;
    }
    const int framesToCount = 200;
    if ((g_curFrameNum % framesToCount) == 0) {
      struct _timeb stop;
      _ftime( &stop );
      float timediff = (stop.time - start.time) + (stop.millitm - start.millitm) / 1000.0;
      printf("Update rate: %f frames/second\n", framesToCount / timediff);
      _ftime( &start );
    }
}


// Choose the vertex and fragment profiles to use.  We need CG_PROFILE_VP30 and
// CG_PROFILE_FP30, respectively.

static void ChooseProfiles()
{
    // Make sure that the appropriate profiles are available on the
    // user's system.
    // try ARBVP1
    if (cgGLIsProfileSupported(CG_PROFILE_ARBVP1))
        vertexProfile = CG_PROFILE_ARBVP1;
    else {
        fprintf(stderr, "ARBVP1 vertex profile not supported on this system.\n");
        exit(1);
    }

    // try FP30
    if (cgGLIsProfileSupported(CG_PROFILE_FP30))
        fragmentProfile = CG_PROFILE_FP30;
    else {
        fprintf(stderr, "fp30 fragment profile not supported on this system.\n");
        exit(1);
    }
}


static void LoadCgPrograms()
{
    assert(cgIsContext(context));

    // Load and compile the vertex program from demo_vert.cg; hold on to the
    // handle to it that is returned.
    vertexProgram = cgCreateProgramFromFile(context, CG_SOURCE, g_vertex_program_name,
                                            vertexProfile, NULL, NULL);
    if (!cgIsProgramCompiled(vertexProgram))
        cgCompileProgram(vertexProgram);

    // Enable the appropriate vertex profile and load the vertex program.
    cgGLEnableProfile(vertexProfile);
    cgGLLoadProgram(vertexProgram);

    // And similarly set things up for the sphere fragment program.
    fragmentProgram1 = cgCreateProgramFromFile(context, CG_SOURCE, g_fragment_program1_name,
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram1))
        cgCompileProgram(fragmentProgram1);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram1);

    // And similarly set things up for the louverblind fragment program.
    fragmentProgram2 = cgCreateProgramFromFile(context, CG_SOURCE, g_fragment_program2_name,
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram2))
        cgCompileProgram(fragmentProgram2);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram2);

    // And similarly set things up for the cone fragment program.
    fragmentProgram3 = cgCreateProgramFromFile(context, CG_SOURCE, g_fragment_program3_name,
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram3))
        cgCompileProgram(fragmentProgram3);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram3);

    // And similarly set things up for the nested spheres fragment program.
    fragmentProgram4 = cgCreateProgramFromFile(context, CG_SOURCE, g_fragment_program4_name,
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram4))
        cgCompileProgram(fragmentProgram4);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram4);
}

inline void Cube(const float x, const float y, const float z)
{
    BindVertexParameters();

    glBegin(GL_QUADS);
      // Bottom (-Z) for the box
      glNormal3f(0,0,-1);
      glVertex3f(-x, y,-z);
      glVertex3f( x, y,-z);
      glVertex3f( x,-y,-z);
      glVertex3f(-x,-y,-z);

      // Top (+Z) for the box
      glNormal3f(0,0,1);
      glVertex3f(-x,-y, z);
      glVertex3f( x,-y, z);
      glVertex3f( x, y, z);
      glVertex3f(-x, y, z);

      // Left (-X) for the box
      glNormal3f(-1,0,0);
      glVertex3f(-x, -y, z);
      glVertex3f(-x,  y, z);
      glVertex3f(-x,  y,-z);
      glVertex3f(-x, -y,-z);

      // Right (+X) for the box
      glNormal3f(1,0,0);
      glVertex3f( x, -y,-z);
      glVertex3f( x,  y,-z);
      glVertex3f( x,  y, z);
      glVertex3f( x, -y, z);

      // front (-Y) for the box
      glNormal3f(0,-1,0);
      glVertex3f(-x, -y, -z);
      glVertex3f( x, -y, -z);
      glVertex3f( x, -y,  z);
      glVertex3f(-x, -y,  z);

      // back (+Y) for the box
      glNormal3f(0,1,0);
      glVertex3f(-x,  y,  z);
      glVertex3f( x,  y,  z);
      glVertex3f( x,  y, -z);
      glVertex3f(-x,  y, -z);
    glEnd();
}

// Make our rocket ship!
static void DrawRocket()
{
  // Set the texture map for the rocket ship.
#ifdef	TEXTURE
    cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram1, "Texture"), g_rocketTexture);
    cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram1, "Texture"));
#endif

  cgGLBindProgram(fragmentProgram1);

  // The main body of the rocket
  glColor3f(1,0.2,0.2);
  glPushMatrix();
    glScalef(1, 1, 7);  
    glRotatef(90.1, 1,0,0);   // Get the texture coordinates lined up the way we want them.
    Cube(1,1,1);
  glPopMatrix();

  glColor3f(0,0.2,0.2);
  // The cockpit bubble
  glPushMatrix();
    glTranslatef(0, 0.5, 5.7);
    glScalef(0.2,0.2,0.4);
    Cube(1,1,1);
  glPopMatrix();

  // The three engines are cones, with little backwards-cones on the ends for caps.
#define enginebacks
  cgGLBindProgram(fragmentProgram3);
  glColor3f(0.2,1,0.2);
  glPushMatrix();
    glRotatef(0,  0,0,1);
    glTranslatef(0, 0.55, -6);
    glScalef(0.5,0.5,2);
    glRotatef(100, 1,0,0);   // Aim them out the back.
    Cube(1,1,1);
#ifdef enginebacks
    glColor3f(1,0.2,0.2);
    cgGLBindProgram(fragmentProgram1);
    glTranslatef(0, -1, 0);
    glRotatef(180, 1,0,0);   // Invert for backwards cone
    glScalef(1,0.05,1);
    Cube(1,1,1);
    cgGLBindProgram(fragmentProgram3);
    glColor3f(0.2,1,0.2);
#endif
  glPopMatrix();
  glPushMatrix();
    glRotatef(120,  0,0,1);
    glTranslatef(0, 0.55, -6);
    glScalef(0.5,0.5,2);
    glRotatef(100, 1,0,0);   // Aim them out the back.
    Cube(1,1,1);
#ifdef enginebacks
    glColor3f(1,0.2,0.2);
    cgGLBindProgram(fragmentProgram1);
    glTranslatef(0, -1, 0);
    glRotatef(180, 1,0,0);   // Invert for backwards cone
    glScalef(1,0.05,1);
    Cube(1,1,1);
    cgGLBindProgram(fragmentProgram3);
    glColor3f(0.2,1,0.2);
#endif
  glPopMatrix();
  glPushMatrix();
    glRotatef(-120,  0,0,1);
    glTranslatef(0, 0.55, -6);
    glScalef(0.5,0.5,2);
    glRotatef(100, 1,0,0);   // Aim them out the back.
    Cube(1,1,1);
#ifdef enginebacks
    glColor3f(1,0.2,0.2);
    cgGLBindProgram(fragmentProgram1);
    glTranslatef(0, -1, 0);
    glRotatef(180, 1,0,0);   // Invert for backwards cone
    glScalef(1,0.05,1);
    Cube(1,1,1);
    glColor3f(0.2,1,0.2);
#endif
   glPopMatrix();
}

// Make our Earth
static void DrawEarth()
{
  // Now bind uniform parameters to fragment program
#ifdef TEXTURE
  cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram1, "Texture"), g_earthTexture);
  cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram1, "Texture"));
#endif
  glColor3f(0.2,0.2,0.8);
  cgGLBindProgram(fragmentProgram1);
  glPushMatrix();
    glScalef(5,5,5);
    Cube(1,1,1);
  glPopMatrix();
}

// Make our Snow-man
static void DrawSnowman()
{
  // Now bind uniform parameters to fragment program
#ifdef TEXTURE
  cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram1, "Texture"), g_snowTexture);
  cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram1, "Texture"));
#endif
  glColor3f(0.2,0.2,0.8);
  cgGLBindProgram(fragmentProgram1);
  glPushMatrix();
    glScalef(5,5,5);
    Cube(1,1,1);
  glPopMatrix();
}

// Make our bouncing ball
static void DrawBouncingBall()
{
  glColor3f(0.8,0.2,0.8);
  glPushMatrix();

  // Now bind uniform parameters to fragment program
#ifdef TEXTURE
  cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram1, "Texture"), g_ballTexture);
  cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram1, "Texture"));
#endif

  // Figure out the fraction of a unit trajectory it is on, where
  // this fraction wraps back around to zero.
  float unitTime = g_curFrameNum/150.0 - floor(g_curFrameNum/150.0);

  // Figure out where it is on a parabolic trajectory that goes up and
  // back down between 0 and 1
  float unitY = 1 - ( 2*(unitTime-0.5) ) * ( 2*(unitTime-0.5) );

  // Translate up and down to match this trajectory
  float Y = 3 * unitY - 7.5;
  glTranslatef(0, Y, 0);

  // Squash the ball in Y at the bottom of its trajectory,
  // preserving volume be stretching in X and Z.  The squash
  // should keep the bottom of the ball steady on the floor,
  // where the floor is located at Y = -8.
  if (Y < -7) {
    float belowFloor = 7 + Y;
    float squashFactorY = 1 - belowFloor;
    glScalef( sqrt(squashFactorY), 1/squashFactorY, sqrt(squashFactorY) );
  }

  cgGLBindProgram(fragmentProgram1);
  Cube(1,1,1);

  glPopMatrix();
}

// Make our earth-orbitted mobile from three rocket ships and Earth
static void DrawMobile()
{
  glTranslatef(-5, 5, 0);
  glScalef(0.5,0.5,0.5);

  // Make the mobile rotate a bit each frame
  // is called.
  glRotatef(g_curFrameNum, 0, 1, 0);
  DrawEarth();

  // Move the spaceship off-center so it looks like it is flying
  glTranslatef(-10, 0, 0);
  DrawRocket();
  glTranslatef(10, 0, 0);
  glRotatef(120, 0, 1, 0);
  glTranslatef(-10, 0, 0);
  DrawRocket();
  glTranslatef(10, 0, 0);
  glRotatef(120, 0, 1, 0);
  glTranslatef(-10, 0, 0);
  DrawRocket();
}

// Make our nested sphere!
static void DrawSphereNest()
{ 
  glColor3f(1.0,1.0,1.0);
  glTranslatef(-5, 5, 0);
  glScalef(0.35,0.35,0.35);

  glRotatef(90.1, 0,0,1);

  glRotatef(g_curFrameNum, 1,0,0);
  cgGLBindProgram(fragmentProgram4);
  Cube(10,10,1);

  glRotatef(90, 1,0,0);
  cgGLBindProgram(fragmentProgram4);
  Cube(10,10,1);
}

// Make our room, standard OpenGL-style.
static void DrawRoom()
{
  // Floor
  glColor3f(0.2,0.2,0.2);
  glBegin(GL_QUADS);
    glVertex3f(-40, -8,-100);
    glVertex3f(-40, -8, 100);
    glVertex3f( 40, -8, 100);
    glVertex3f( 40, -8,-100);
  glEnd();

  // Ceiling
  glColor3f(0.9,0.9,0.9);
  glBegin(GL_QUADS);
    glVertex3f( 40, 25,-100);
    glVertex3f( 40, 25, 100);
    glVertex3f(-40, 25, 100);
    glVertex3f(-40, 25,-100);
  glEnd();

  // Right Wall
  glColor3f(0.5,0.5,0.2);
  glBegin(GL_QUADS);
    glVertex3f( 40, 25,-100);
    glVertex3f( 40, -8,-100);
    glVertex3f( 40, -8, 100);
    glVertex3f( 40, 25, 100);
  glEnd();

  // Left Wall
  glColor3f(0.5,0.5,0.2);
  glBegin(GL_QUADS);
    glVertex3f(-40, 25, 100);
    glVertex3f(-40, -8, 100);
    glVertex3f(-40, -8,-100);
    glVertex3f(-40, 25,-100);
  glEnd();

  // Far Wall
  //XXX Need to cut a hole in the wall for the window.
  glColor3f(0.3,0.3,0.1);
  glBegin(GL_QUADS);
    glVertex3f(-40, 25,-100);
    glVertex3f(-40, 18,-100);
    glVertex3f( 40, 18,-100);
    glVertex3f( 40, 25,-100);
  glEnd();
  glBegin(GL_QUADS);
    glVertex3f(-40, -1,-100);
    glVertex3f(-40, -8,-100);
    glVertex3f( 40, -8,-100);
    glVertex3f( 40, -1,-100);
  glEnd();
  glBegin(GL_QUADS);
    glVertex3f(-40, 25,-100);
    glVertex3f(-40, -8,-100);
    glVertex3f( 10, -8,-100);
    glVertex3f( 10, 25,-100);
  glEnd();
  glBegin(GL_QUADS);
    glVertex3f( 30, 25,-100);
    glVertex3f( 30, -8,-100);
    glVertex3f( 50, -8,-100);
    glVertex3f( 50, 25,-100);
  glEnd();

  // Blue sky backdrop outside window
  glColor3f(0.3,0.3,1);
  glBegin(GL_QUADS);
    glVertex3f(-100,  100,-150);
    glVertex3f(-100, -100,-150);
    glVertex3f( 100, -100,-150);
    glVertex3f( 100,  100,-150);
  glEnd();
}


static void keyboard(unsigned char key, int x, int y)
{
    if (key == 'q' || key == 'Q' || key == 27 /* escape */) {
        cgDestroyContext(context);
        exit(0);
    }
    if (key == 'f' || key == 'F') {
        g_stopMotion = !g_stopMotion;
    }
    if (key == 'p' || key == 'P') {
        g_showPolies = !g_showPolies;
    }
    if (key == 't' || key == 'T') {
        g_toggle = !g_toggle;
    }
    if (key == 'm' || key == 'M') {
        g_whichMobile++;
    }
}

static void handleCgError() 
{
    fprintf(stderr, "Cg error: %s\n", cgGetErrorString(cgGetError()));
    exit(1);
}
