/*********************************************************************NVMH3****

Copyright NVIDIA Corporation 2002
TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
*AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.


Comments:
  This code is based on the demo.cpp example program from NVIDIA described by the above
  Copyright.  It has been extended by Russ Taylor to handle vertex-plus-fragment
  code raytracing of Non-Polygonal objects, in particular implicit surfaces, in
  particular quadratics, in particular spheres.

******************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef WIN32
# include <windows.h>
#endif
#ifdef __APPLE__
# include <OpenGL/gl.h>
# include <OpenGL/glext.h>
# include <GLUT/glut.h>
#else
# include <GL/gl.h>
# include <GL/glut.h>
# include <GL/glext.h>
#endif
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#ifndef CWD
# define CWD ""
#endif

// This should be defined (here or on a config file) if the vertex
// program is to compute the values needed by the fragment program
// to calculate the proper Z.  If this is not done, the Z will be
// good enough if there are no other primitives intersecting the
// polygons bounding the hull of the NPO, but will not be correct
// for intersection tests with the NPO itself.
#define NPO_EXACT_Z


// Global variables: hold the Cg context that we're storing our programs
// in as well as handles to the vertex and fragment program used in this
// demo.

CGcontext context;
CGprogram vertexProgram, fragmentProgram;
CGprofile vertexProfile, fragmentProfile;


// More globals
bool  g_drawPolies = false;	// Should we draw the polygon stand-in geometry? XXX Fails.
bool  g_drawWireframe = false;  // Should we draw in wire-frame? XXX Fails.
bool  g_freezeMotion = false;	// Should we stop the motion of the object now?
const double g_hitherDist = 0.1;
const double g_yonDist = 100.0;

// Forward declarations of functions

static void keyboard(unsigned char key, int x, int y);
static void display();
static void handleCgError();
static void DrawGeometry();
static void LoadTextures();
static void LoadCgPrograms();
static void ChooseProfiles();

///////////////////////////////////////////////////////////////////////////

// Main program; do basic GLUT and Cg setup, but leave most of the work
// to the display() function.

int main(int argc, char *argv[]) {
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_ALPHA | GLUT_DEPTH);
    glutInitWindowSize(512, 512);
    glutCreateWindow("RayFragments");

    glutKeyboardFunc(keyboard);
    glutDisplayFunc(display);

    // Basic Cg setup; register a callback function for any errors
    // and create an initial context
    cgSetErrorCallback(handleCgError);
    context = cgCreateContext();

    // and all the rest happens in the display function...
    glutMainLoop();

    return 0;
}


// display callback function

static void display() 
{
    static int curFrameNum = 0;

    static bool haveLoadedPrograms = false;
    if (!haveLoadedPrograms) {
	// Do one-time setup only once; setup Cg programs and textures
	// and set up OpenGL state.
        ChooseProfiles();
        LoadCgPrograms();
        LoadTextures();

        glEnable(GL_DEPTH_TEST);
        haveLoadedPrograms = true;
    }

    // The usual OpenGL stuff to clear the screen and set up viewing.
    glClearColor(.25, .25, .25, 1.);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(30., 1.0, g_hitherDist, g_yonDist);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(0, 0, 8, 0, 0, 0, 0, 1, 0);

    // Make the object rotate a bit each time the display function
    // is called
    glRotatef(curFrameNum, 0, 1, 0);
    //glRotatef(curFrameNum*1.3, 1, 0, 1);
    //glScalef(1 + 0.5*sin(curFrameNum/12.0),1,1);
    glScalef(1,1,1);

    // Now make sure that the vertex and fragment programs, loaded
    // in LoadCgPrograms() are bound.
    if (g_drawPolies) {
      // Causes all bound profiles to stop for these profiles, dropping
      // back to polygon rendering.
      cgGLDisableProfile(vertexProfile);
      cgGLDisableProfile(fragmentProfile);
    } else {
      cgGLEnableProfile(vertexProfile);
      cgGLEnableProfile(fragmentProfile);
      cgGLBindProgram(vertexProgram);
      cgGLBindProgram(fragmentProgram);
    }

    // Bind uniform parameters to vertex shader
    cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram, "ModelViewProj"),
                                CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                CG_GL_MATRIX_IDENTITY);
    cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram, "ModelViewI"),
                                CG_GL_MODELVIEW_MATRIX,
                                CG_GL_MATRIX_INVERSE);

    // Now bind uniform parameters to vertex and fragment shaders
#ifdef NPO_EXACT_Z
    cgGLSetParameter1f(cgGetNamedParameter(vertexProgram, "Zhither"), g_hitherDist);
    cgGLSetParameter1f(cgGetNamedParameter(vertexProgram, "Zyon"), g_yonDist);
#endif

/*XXX
    // We can also go ahead and bind varying parameters to vertex shader
    // that we just want to have the same value for all vertices.  The
    // vertex shader could be modified so that these were uniform for
    // better efficiency, but this gives us flexibility for the future.
    float Kd[3] = { .7, .2, .2 }, Ks[3] = { .9, .9, .9 };
    cgGLSetParameter3fv(cgGetNamedParameter(vertexProgram, "diffuse"), Kd);
    cgGLSetParameter3fv(cgGetNamedParameter(vertexProgram, "specular"), Ks);

    // Now bind uniform parameters to fragment shader
    float lightPos[3] = { 3, 2, -3 };
    cgGLSetParameter3fv(cgGetNamedParameter(fragmentProgram, "Plight"), lightPos);
    float lightColor[3] = { 1, 1, 1 };
    cgGLSetParameter3fv(cgGetNamedParameter(fragmentProgram, "lightColor"), 
                        lightColor);
    cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram, "shininess"), 40);

    // And finally, enable the approprate texture for fragment shader; the
    // texture was originally set up in LoadTextures().
    cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram,
						   "diffuseMap"));
*/
    // And go ahead and draw the scene geometry
    DrawGeometry();
    // Disable the texture now that we're done with it.
/*XXX
    cgGLDisableTextureParameter(cgGetNamedParameter(fragmentProgram,
						    "diffuseMap"));
*/

    glutSwapBuffers();
    if (!g_freezeMotion) {
      ++curFrameNum;
    }
    // Force another display(), so that the object keeps moving
    glutPostRedisplay();
}


// Choose the vertex and fragment profiles to use.  We need CG_PROFILE_VP30 and
// CG_PROFILE_FP30, respectively.

static void ChooseProfiles()
{
    // Make sure that the appropriate profiles are available on the
    // user's system.
    // try VP30
    if (cgGLIsProfileSupported(CG_PROFILE_VP30))
        vertexProfile = CG_PROFILE_VP30;
    else {
        fprintf(stderr, "vp30 vertex profile not supported on this system.\n");
        exit(1);
    }

    // try FP30
    if (cgGLIsProfileSupported(CG_PROFILE_FP30))
        fragmentProfile = CG_PROFILE_FP30;
    else {
        fprintf(stderr, "fp30 fragment profile not supported on this system.\n");
        exit(1);
    }
}


static void LoadCgPrograms()
{
    assert(cgIsContext(context));

    // Load and compile the vertex program from demo_vert.cg; hold on to the
    // handle to it that is returned.
    vertexProgram = cgCreateProgramFromFile(context, CG_SOURCE, CWD "NPO_vertex.cg",
                                            vertexProfile, NULL, NULL);
    if (!cgIsProgramCompiled(vertexProgram))
        cgCompileProgram(vertexProgram);

    // Enable the appropriate vertex profile and load the vertex program.
    cgGLEnableProfile(vertexProfile);
    cgGLLoadProgram(vertexProgram);

    // And similarly set things up for the fragment program.
    fragmentProgram = cgCreateProgramFromFile(context, CG_SOURCE, CWD "NPO_sphere_fragment.cg",
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram))
        cgCompileProgram(fragmentProgram);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram);
}

static void LoadTextures()
{
    // There is only one texture needed here--we'll set up a basic
    // checkerboard--which is used to modulate the diffuse channel in the
    // fragment shader.
    GLuint handle;
    glGenTextures(1, &handle);

    // Basic OpenGL texture state setup
    glBindTexture(GL_TEXTURE_2D, handle);
    glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    // Fill in the texture map.
#define RES 512
    float *data = new float[RES*RES*4];
    float *dp = data;
    for (int i = 0; i < RES; ++i) {
        for (int j = 0; j < RES; ++j) {
            if ((i/32+j/32) % 2) {
                *dp++ = .7;
                *dp++ = .7;
                *dp++ = .7;
            }
            else {
                *dp++ = .1;
                *dp++ = .1;
                *dp++ = .1;
            }
	    *dp++ = 1.;
        }
    }

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, RES, RES, 0, GL_RGBA, GL_FLOAT, data);
    delete[] data;

    // Tell Cg which texture handle should be associated with the sampler2D
    // parameter to the fragment shader.
/*XXX Put the texture map name back in later.
    cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram, "diffuseMap"),
                            handle);
*/
}

// Geometry creation and drawing function; we'll just draw a cube.

static void DrawGeometry()
{
  glCullFace(GL_BACK);
  glEnable(GL_CULL_FACE);

  glBegin(GL_QUADS);
    // Bottom (-Z) for the box
    glVertex3f(-1, 1,-1);
    glVertex3f( 1, 1,-1);
    glVertex3f( 1,-1,-1);
    glVertex3f(-1,-1,-1);

    // Top (+Z) for the box
    glVertex3f(-1,-1, 1);
    glVertex3f( 1,-1, 1);
    glVertex3f( 1, 1, 1);
    glVertex3f(-1, 1, 1);

/*XXX
    // Left (-X) for the box
    glVertex3f(-1, -1, 1);
    glVertex3f(-1,  1, 1);
    glVertex3f(-1,  1,-1);
    glVertex3f(-1, -1,-1);

    // Right (+X) for the box
    glVertex3f( 1, -1,-1);
    glVertex3f( 1,  1,-1);
    glVertex3f( 1,  1, 1);
    glVertex3f( 1, -1, 1);

    // front (-Y) for the box
    glVertex3f(-1, -1, -1);
    glVertex3f( 1, -1, -1);
    glVertex3f( 1, -1,  1);
    glVertex3f(-1, -1,  1);

    // back (+Y) for the box
    glVertex3f(-1,  1,  1);
    glVertex3f( 1,  1,  1);
    glVertex3f( 1,  1, -1);
    glVertex3f(-1,  1, -1);
*/
  glEnd();
}

static void keyboard(unsigned char key, int x, int y)
{
    if (key == 'q' || key == 'Q' || key == 27 /* escape */) {
        cgDestroyContext(context);
        exit(0);
    }
    if (key == 'p' || key == 'P') {
	g_drawPolies = !g_drawPolies;
    }
    if (key == 'w' || key == 'W') {
	g_drawWireframe = !g_drawWireframe;
    }
    if (key == 'f' || key == 'F') {
	g_freezeMotion = !g_freezeMotion;
    }
}

static void handleCgError() 
{
    fprintf(stderr, "Cg error: %s\n", cgGetErrorString(cgGetError()));
    exit(1);
}
