/******************************************************************************

Comments:
  This code is based on the demo.cpp example program from NVIDIA described by the above
  Copyright.  It has been extended by Russ Taylor to handle vertex-plus-fragment
  code raytracing of Non-Polygonal objects, in particular implicit surfaces, in
  particular quadratics, in particular spheres.

  XXX We'd like the current openGL matrix values to be available in a fragment
  program so that display lists can be sent in and the matrix not have to be
  re-bound before each object.

******************************************************************************/

/******************************************************
 * Speed tests:

  See below for why Initial try was bogus:
   Initial try, 470K sets of 6 quads/second (no spheres, bounding geometry including xforms).
		32K spheres/second.  (~30 triangles/sphere? hmm...)
   Inlining BindVertexParameters got up to 34K/sec.
   Passing only float3s from vertex to fragment didn't help.
   Not setting color in fragment program when sphere is missed didn't help.
   Making the fragment shader only set color and depth didn't help.
   * --> Not fragment-program bound!
   Only binding the vertex parameters 1/frame got us to 76K/sec (but recall, sphere always centered).
     Only doing the string-based parameter lookup the first time didn't help much at all.
   * --> Not bound by binding the parameters!
   Avoiding the back-conversion of the eye position to object space didn't help.
   * --> Indicates maybe not vertex program-sensitive
   Using the demonstration vertex and fragment shaders instead of mine didn't help hardly at all.
     Switching the vertex profile to ARBVP1 hardly helped at all (even for demo programs).
     Switching the fragment profile to ARBFP1 hardly helped at all (even for demo programs).
   * --> Not a problem with shaders more complicated than usual?
   * --> Not a problem with what profile we are using?
   CPU meter shows 100% CPU usage with only a tiny bit of system time
   * --> Something in user-level code in the program is our bottleneck.
   Moving the profile-enabling code out of the display loop only helped a teeny bit.
   Putting the display of the quads for a cube into a display list didn't help.
      I think it made both the sphere code and the quad code slower...
   Scaling the whole sphere array by 10 on a side didn't change the speed.
     Although it slowed down the drawing-quads case!
   * --> Even though most of the spheres were outside the clipping volume in this case!
   Making the window larger with the grid 10x the size made things half as fast
     But with the grid at the original size, the larger image was fast again.
   Putting a 1ms sleep into the display() routine dropped the CPU to 85% and the update rate to 32K/sec.
   XXX Are vertex arrays going to be my friend?  Maybe, but not the bottleneck, given what comes next...
   Doing everything but not drawing the geometry get us about 63K
   * --> This means that the bottleneck is not in the vertex or fragment programs, which don't run then!
     The quads routine goes to 3.7M when drawing is turned off.
   * --> This only leaves the setting of the vertex parameters and the frame setup/teardown...
   Taking out the loop of translate, setup, draw bumps us to the frame rate limit.
   Taking out the translates has minimal effect.
   Taking out the binding of the parameters bumps us to framerate, but still has high load.
   Switching to exact sphere intersections drops it to 33.3K (remember that we're sending more state).

  More appropriate test:
   The initial try was not using lit surfaces or geometry that approximates a sphere.
   When we turned on lighting and used glutSolidSphere(1.0, 4, 4) to approximate the
   spheres (a poor approximation, especially for lighting), the speed was about the
   same for the polygon case and the vertex-program case.  Display lists seem to make
   the polygon spheres much faster (though making the CG ones slower), which meant
   that glutSolidSphere(1.0, 6, 5) runs about the same speed as the CG program on my
   home machine.  This is still a poor approximation, especially for lighting.  Note
   that the light is from the viewing direction for this test because lighting direction
   is not specified in the current CG program.
   Surprising to me, putting all of the polygon spheres into one big display list
   (including the transformations) made things SLOWER than having just the sphere itself
   in the display list and calling multiple draws on it with different transforms sent
   in between.
   The 700x700 image version of the current exact CG code goes at just over 33,000 spheres/second.
   The polygon version with 6x6 tesselation goes at 26,700 straight up and at 32,200 with display lists.

   * -> Removing ModelView matrix send (not needed) got CG to 37,500.
   * -> Using tightly-fit tetrahedra rather than cubes dropped CG to 24,200 (no longer CPU bound)
	(Drawing more pixels in the surround, backface culling helped quads?)
 *
 ******************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef WIN32
# include <windows.h>
#endif
#ifdef __APPLE__
# include <OpenGL/gl.h>
# include <OpenGL/glext.h>
# include <GLUT/glut.h>
#else
# include <GL/gl.h>
# include <GL/glut.h>
# include <GL/glext.h>
# include <sys/types.h>
# include <sys/timeb.h>
#endif
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#ifndef CWD
# define CWD ""
#endif

// Turn this off to test just the quad update rate, for comparison.
const g_useCG = true;
const int g_numSpheres = 2000;
const int g_sphereTessX = 6, g_sphereTessY = 6;
//#define	USE_DLISTS

// Define this to use tetrahedra rather than cubes
//#define	USE_TETRAHEDRA
const float g_tetcoord = 1.7; // Corner of coordinate as one point of a cube

// Turn this on to jitter within the lattice, so there is overlap
const g_jitterPos = true;

// Display lists slow down BOTH quad and sphere display!

const char g_vertex_program_name[] = CWD "NPO_vertex.cg";
const char g_fragment_program_name[] = CWD "NPO_sphere_fragment.cg";

// Global variables: hold the Cg context that we're storing our programs
// in as well as handles to the vertex and fragment program used in this
// demo.

CGcontext context;
CGprogram vertexProgram, fragmentProgram;
CGprofile vertexProfile, fragmentProfile;

// More globals
const double g_hitherDist = 50;
const double g_yonDist = 150.0;

// Global list of steps to get from one sphere center to the next.
typedef	float CENTERPOINT[3];
CENTERPOINT g_offsets[g_numSpheres];

// Forward declarations of functions

static void keyboard(unsigned char key, int x, int y);
static void display();
static void handleCgError();
static void DrawGeometry();
static void LoadCgPrograms();
static void ChooseProfiles();

///////////////////////////////////////////////////////////////////////////
// Fill the point list with offsets, so that each point stays within a
// cube but we don't need to push and pop matrix states.

static void ComputeSphereOffsets()
{
    // Make a bunch of spheres on random locations in a uniform
    // lattice that has a separation of 2 in X, Y, and Z and that
    // has sides going from -20 to 20.  This must be done using
    // translation matrices so that the vertex coordinates in
    // object space maintain their surround of the sphere.  Some
    // of these may be overlapping.

    int i;
    srand(1); // Make the same set each time.
    CENTERPOINT origin = { 0, 0, 0 };
    CENTERPOINT lastPoint;
    g_offsets[0][0] = lastPoint[0] = origin[0];
    g_offsets[0][1] = lastPoint[1] = origin[1];
    g_offsets[0][2] = lastPoint[2] = origin[2];
    for (i = 1; i < g_numSpheres; i++) {
      CENTERPOINT newPoint = { ((rand() % 21) - 10) * 2,
		    	       ((rand() % 21) - 10) * 2,
			       ((rand() % 21) - 10) * 2 };

      // Jitter to non-lattice point if this mode is set
      if (g_jitterPos) {
	newPoint[0] += ((rand() % 11) - 5.0) / 2.5;
	newPoint[1] += ((rand() % 11) - 5.0) / 2.5;
	newPoint[2] += ((rand() % 11) - 5.0) / 2.5;
      }

      g_offsets[i][0] = newPoint[0] - lastPoint[0];
      g_offsets[i][1] = newPoint[1] - lastPoint[1];
      g_offsets[i][2] = newPoint[2] - lastPoint[2];

      lastPoint[0] = newPoint[0];
      lastPoint[1] = newPoint[1];
      lastPoint[2] = newPoint[2];
    }
}

// NOTE: When these get bound, they set the modelview matrix and others
// to the current one on the stack -- this makes any modeling transform
// that happens in the display list of the object have no effect!!!!!
// This means that we can't do vertex-list primitives and use the
// model transform.
inline void BindVertexParameters()
{
  static bool initialized = false;
  static CGparameter ModelViewProj, ModelViewI, ModelViewIT, ModelView;

  if (!initialized) {
    ModelViewProj = cgGetNamedParameter(vertexProgram, "ModelViewProj");
    ModelViewI = cgGetNamedParameter(vertexProgram, "ModelViewI");
    initialized = true;
  }

  cgGLSetStateMatrixParameter(ModelViewProj,
			      CG_GL_MODELVIEW_PROJECTION_MATRIX,
			      CG_GL_MATRIX_IDENTITY);
  cgGLSetStateMatrixParameter(ModelViewI,
			      CG_GL_MODELVIEW_MATRIX,
			      CG_GL_MATRIX_INVERSE);
}

///////////////////////////////////////////////////////////////////////////
// Main program; do basic GLUT and Cg setup, but leave most of the work
// to the display() function.

int main(int argc, char *argv[]) {
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_ALPHA | GLUT_DEPTH);
    glutInitWindowSize(700, 700);
    glutCreateWindow("RayFragments");

    glutKeyboardFunc(keyboard);
    glutDisplayFunc(display);

    // Basic Cg setup; register a callback function for any errors
    // and create an initial context
    cgSetErrorCallback(handleCgError);
    context = cgCreateContext();

    // Set up Cg programs
    static bool haveLoadedPrograms = false;
    if (!haveLoadedPrograms) {
	// Do one-time setup only once; setup Cg programs and textures
	// and set up OpenGL state.
        ChooseProfiles();
        LoadCgPrograms();

        glEnable(GL_DEPTH_TEST);
        haveLoadedPrograms = true;
    }

    if (g_useCG) {
      // Now make sure that the vertex and fragment programs, loaded
      // in LoadCgPrograms() are bound.
      cgGLEnableProfile(vertexProfile);
      cgGLEnableProfile(fragmentProfile);
      cgGLBindProgram(vertexProgram);
      cgGLBindProgram(fragmentProgram);

    // Now bind uniform parameters to vertex program
      cgGLSetParameter1f(cgGetNamedParameter(vertexProgram, "Zhither"), g_hitherDist);
      cgGLSetParameter1f(cgGetNamedParameter(vertexProgram, "Zyon"), g_yonDist);

      // Now bind uniform parameters to fragment programs
      cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram, "Zhither"), g_hitherDist);
      cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram, "Zyon"), g_yonDist);
    } else {
      // Set up lighting for the polygonal spheres
      GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
      GLfloat mat_diffuse[] = { 0.6, 0.1, 0.6, 1.0 };
      GLfloat mat_shininess[] = { 50.0 };
      GLfloat light_position[] = { 0.0, 0.0, 1.0, 0.0 };
      glShadeModel (GL_SMOOTH);
      glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
      glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_diffuse);
      glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
      glLightfv(GL_LIGHT0, GL_POSITION, light_position);
      glEnable(GL_LIGHTING);
      glEnable(GL_LIGHT0);
      glEnable(GL_LIGHTING);
    }

    // Compute the sphere-center coordinates, and the translations
    // from the center of each to get to the center of the next.
    ComputeSphereOffsets();

    // and all the rest happens in the display function...
    glutMainLoop();

    return 0;
}

// display callback function

static void display() 
{
    static int curFrameNum = 0;
    static struct _timeb start;

    // The usual OpenGL stuff to clear the screen and set up viewing.
    glClearColor(.25, .25, .25, 1.);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(40., 1.0, g_hitherDist, g_yonDist);

    glMatrixMode(GL_MODELVIEW);

    glLoadIdentity();
    gluLookAt(0, 0, 100, 0, 0, 0, 0, 1, 0);

    // Make the object rotate a bit each time the display function
    // is called
    glRotatef(curFrameNum, 0, 1, 0);
    //glScalef(10,10,10);
    //glRotatef(curFrameNum*1.3, 1, 0, 1);
    //glScalef(1 + 0.5*sin(curFrameNum/12.0),1,1);
    //glScalef(1,1,1);

    // Set things up so that backface culling is on.
    glCullFace(GL_BACK);
    glEnable(GL_CULL_FACE);

    // Draw spheres whose centers are displaced one to the other by
    // the stored array of offsets.  Send the new matrices to the
    // vertex program for each sphere.

    int i;
    for (i = 0; i < g_numSpheres; i++) {
      glTranslatef(g_offsets[i][0], g_offsets[i][1], g_offsets[i][2]);
      if (g_useCG) {
	BindVertexParameters();
      }
      DrawGeometry();
    }

    glutSwapBuffers();

    // Handle timing information
    if (curFrameNum == 0) {
      _ftime( &start );
    }
    ++curFrameNum;
    const int framesToCount = 90;
    if ((curFrameNum % framesToCount) == 0) {
      struct _timeb stop;
      _ftime( &stop );
      float timediff = (stop.time - start.time) + (stop.millitm - start.millitm) / 1000.0;
      printf("Update rate: %f spheres/second:\n", framesToCount * g_numSpheres / timediff);
      printf("             %f frames/second)\n", framesToCount / timediff);
      printf("             %d spheres\n", g_numSpheres);
      _ftime( &start );
    }

    // Force another display(), so that the objects keep moving
    glutPostRedisplay();
}


// Choose the vertex and fragment profiles to use.  We need CG_PROFILE_VP30 and
// CG_PROFILE_FP30, respectively.

static void ChooseProfiles()
{
    // Make sure that the appropriate profiles are available on the
    // user's system.
    // try VP30
    if (cgGLIsProfileSupported(CG_PROFILE_ARBVP1))
        vertexProfile = CG_PROFILE_ARBVP1;
    else {
        fprintf(stderr, "vp30 vertex profile not supported on this system.\n");
        exit(1);
    }

    // try FP30
    if (cgGLIsProfileSupported(CG_PROFILE_FP30))
        fragmentProfile = CG_PROFILE_FP30;
    else {
        fprintf(stderr, "fp30 fragment profile not supported on this system.\n");
        exit(1);
    }
}


static void LoadCgPrograms()
{
  if (g_useCG) {
    assert(cgIsContext(context));

    // Load and compile the vertex program from demo_vert.cg; hold on to the
    // handle to it that is returned.
    vertexProgram = cgCreateProgramFromFile(context, CG_SOURCE, g_vertex_program_name,
                                            vertexProfile, NULL, NULL);
    if (!cgIsProgramCompiled(vertexProgram))
        cgCompileProgram(vertexProgram);

    // Enable the appropriate vertex profile and load the vertex program.
    cgGLEnableProfile(vertexProfile);
    cgGLLoadProgram(vertexProgram);

    // And similarly set things up for the fragment program.
    fragmentProgram = cgCreateProgramFromFile(context, CG_SOURCE, g_fragment_program_name,
                                              fragmentProfile, NULL, NULL);
    if (!cgIsProgramCompiled(fragmentProgram))
        cgCompileProgram(fragmentProgram);

    cgGLEnableProfile(fragmentProfile);
    cgGLLoadProgram(fragmentProgram);
  }
}

// Geometry creation and drawing function; we'll just draw a cube.

static void DrawGeometry()
{
#ifdef	USE_DLISTS
  // The first time through, build a display list.
  // Later times, just call the display list.
  static bool	  initialized = false;
  static GLuint	  dList;

  if (initialized) {
    glCallList(dList);
  } else {
    initialized = true;
    dList = glGenLists(1);
    glNewList(dList, GL_COMPILE_AND_EXECUTE);
#endif
    if (g_useCG) {
#ifdef USE_TETRAHEDRA
      glBegin(GL_TRIANGLES);
#define v1 glVertex3f( g_tetcoord, g_tetcoord, g_tetcoord)
#define v2 glVertex3f(-g_tetcoord,-g_tetcoord, g_tetcoord)
#define v3 glVertex3f(-g_tetcoord, g_tetcoord,-g_tetcoord)
#define v4 glVertex3f( g_tetcoord,-g_tetcoord,-g_tetcoord)

	v2; v1; v3; // Correct handedness
	v1; v2; v4; // Correct handedness
	v1; v4; v3; // Correct handedness
	v2; v3; v4; // Correct handedness
      glEnd();
#else
      glBegin(GL_QUADS);
	// Bottom (-Z) for the box
	glVertex3f(-1, 1,-1);
	glVertex3f( 1, 1,-1);
	glVertex3f( 1,-1,-1);
	glVertex3f(-1,-1,-1);

	// Top (+Z) for the box
	glVertex3f(-1,-1, 1);
	glVertex3f( 1,-1, 1);
	glVertex3f( 1, 1, 1);
	glVertex3f(-1, 1, 1);

	// Left (-X) for the box
	glVertex3f(-1, -1, 1);
	glVertex3f(-1,  1, 1);
	glVertex3f(-1,  1,-1);
	glVertex3f(-1, -1,-1);

	// Right (+X) for the box
	glVertex3f( 1, -1,-1);
	glVertex3f( 1,  1,-1);
	glVertex3f( 1,  1, 1);
	glVertex3f( 1, -1, 1);

	// front (-Y) for the box
	glVertex3f(-1, -1, -1);
	glVertex3f( 1, -1, -1);
	glVertex3f( 1, -1,  1);
	glVertex3f(-1, -1,  1);

	// back (+Y) for the box
	glVertex3f(-1,  1,  1);
	glVertex3f( 1,  1,  1);
	glVertex3f( 1,  1, -1);
	glVertex3f(-1,  1, -1);
      glEnd();
#endif
    } else {
      glutSolidSphere(1.0, g_sphereTessX, g_sphereTessY);
    }
#ifdef	USE_DLISTS
    glEndList();
  }
#endif
}

static void keyboard(unsigned char key, int x, int y)
{
    if (key == 'q' || key == 'Q' || key == 27 /* escape */) {
        cgDestroyContext(context);
        exit(0);
    }
}

static void handleCgError() 
{
    fprintf(stderr, "Cg error: %s\n", cgGetErrorString(cgGetError()));
    exit(1);
}
